import itertools
import json
import time
import unittest
from typing import Dict

from locust.dispatch import UsersDispatcher
from locust.runners import WorkerNode
from locust.test.util import clear_all_functools_lru_cache

_TOLERANCE = 0.025


class TestAssignUsersToWorkers(unittest.TestCase):
    def test_assign_users_to_1_worker(self):
        worker_node1 = WorkerNode("1")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1], user_classes_count={"User1": 3, "User2": 3, "User3": 3}, spawn_rate=1
        )
        self.assertDictEqual(
            users_dispatcher._desired_users_assigned_to_workers, {"1": {"User1": 3, "User2": 3, "User3": 3}}
        )

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1], user_classes_count={"User1": 5, "User2": 4, "User3": 2}, spawn_rate=1
        )
        self.assertDictEqual(
            users_dispatcher._desired_users_assigned_to_workers, {"1": {"User1": 5, "User2": 4, "User3": 2}}
        )

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1], user_classes_count={"User1": 1, "User2": 1, "User3": 1}, spawn_rate=1
        )
        self.assertDictEqual(
            users_dispatcher._desired_users_assigned_to_workers, {"1": {"User1": 1, "User2": 1, "User3": 1}}
        )

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1], user_classes_count={"User1": 1, "User2": 1, "User3": 0}, spawn_rate=1
        )
        self.assertDictEqual(
            users_dispatcher._desired_users_assigned_to_workers, {"1": {"User1": 1, "User2": 1, "User3": 0}}
        )

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1], user_classes_count={"User1": 0, "User2": 0, "User3": 0}, spawn_rate=1
        )
        self.assertDictEqual(
            users_dispatcher._desired_users_assigned_to_workers, {"1": {"User1": 0, "User2": 0, "User3": 0}}
        )

    def test_assign_users_to_3_workers(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "1": {"User1": 1, "User2": 1, "User3": 1},
            "2": {"User1": 1, "User2": 1, "User3": 1},
            "3": {"User1": 1, "User2": 1, "User3": 1},
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 5, "User2": 4, "User3": 2},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "1": {"User1": 2, "User2": 1, "User3": 1},
            "2": {"User1": 2, "User2": 1, "User3": 1},
            "3": {"User1": 1, "User2": 2, "User3": 0},
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 1, "User2": 1, "User3": 1},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "1": {"User1": 1, "User2": 0, "User3": 0},
            "2": {"User1": 0, "User2": 1, "User3": 0},
            "3": {"User1": 0, "User2": 0, "User3": 1},
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 1, "User2": 1, "User3": 0},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "1": {"User1": 1, "User2": 0, "User3": 0},
            "2": {"User1": 0, "User2": 1, "User3": 0},
            "3": {"User1": 0, "User2": 0, "User3": 0},
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 0, "User2": 0, "User3": 0},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "1": {"User1": 0, "User2": 0, "User3": 0},
            "2": {"User1": 0, "User2": 0, "User3": 0},
            "3": {"User1": 0, "User2": 0, "User3": 0},
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

    def test_assign_5_users_to_10_workers(self):
        # Prepend "0" to worker name under 10, because workers are sorted in alphabetical orders.
        # It's simply makes the test easier to read, but in reality, the worker "10" would come
        # before worker "1".
        worker_nodes = [WorkerNode("0{}".format(i) if i < 10 else str(i)) for i in range(1, 11)]

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 10, "User2": 5, "User3": 5, "User4": 5, "User5": 5},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "06": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "07": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 11, "User2": 5, "User3": 5, "User4": 5, "User5": 5},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 2, "User2": 1, "User3": 0, "User4": 0, "User5": 1},  # 4 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "06": {"User1": 1, "User2": 0, "User3": 1, "User4": 1, "User5": 0},  # 3 users
            "07": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 11, "User2": 5, "User3": 5, "User4": 5, "User5": 6},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 2, "User2": 1, "User3": 0, "User4": 0, "User5": 1},  # 4 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "06": {"User1": 1, "User2": 0, "User3": 1, "User4": 1, "User5": 0},  # 3 users
            "07": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 11, "User2": 5, "User3": 5, "User4": 6, "User5": 6},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 2, "User2": 1, "User3": 0, "User4": 0, "User5": 1},  # 4 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "06": {"User1": 1, "User2": 0, "User3": 1, "User4": 1, "User5": 0},  # 3 users
            "07": {"User1": 1, "User2": 0, "User3": 0, "User4": 2, "User5": 0},  # 3 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 11, "User2": 5, "User3": 6, "User4": 6, "User5": 6},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 2, "User2": 1, "User3": 0, "User4": 0, "User5": 1},  # 4 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "06": {"User1": 1, "User2": 0, "User3": 1, "User4": 1, "User5": 0},  # 3 users
            "07": {"User1": 1, "User2": 0, "User3": 1, "User4": 1, "User5": 0},  # 3 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 2, "User5": 0},  # 3 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 11, "User2": 6, "User3": 6, "User4": 6, "User5": 6},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 2, "User2": 1, "User3": 0, "User4": 0, "User5": 1},  # 4 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "06": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 3 users
            "07": {"User1": 1, "User2": 0, "User3": 1, "User4": 1, "User5": 0},  # 3 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 2, "User5": 0},  # 3 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 2, "User5": 0},  # 3 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 11, "User2": 6, "User3": 6, "User4": 6, "User5": 7},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 2, "User2": 1, "User3": 0, "User4": 0, "User5": 1},  # 4 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "06": {"User1": 1, "User2": 1, "User3": 1, "User4": 0, "User5": 1},  # 4 users
            "07": {"User1": 1, "User2": 0, "User3": 1, "User4": 1, "User5": 0},  # 3 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 2, "User5": 0},  # 3 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 2, "User5": 0},  # 3 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 1},  # 3 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes,
            user_classes_count={"User1": 11, "User2": 6, "User3": 6, "User4": 6, "User5": 11},
            spawn_rate=1,
        )
        expected_balanced_users = {
            "01": {"User1": 2, "User2": 1, "User3": 1, "User4": 0, "User5": 0},  # 4 users
            "02": {"User1": 1, "User2": 1, "User3": 1, "User4": 1, "User5": 0},  # 4 users
            "03": {"User1": 1, "User2": 1, "User3": 1, "User4": 1, "User5": 0},  # 4 users
            "04": {"User1": 1, "User2": 1, "User3": 1, "User4": 1, "User5": 0},  # 4 users
            "05": {"User1": 1, "User2": 1, "User3": 1, "User4": 1, "User5": 0},  # 4 users
            "06": {"User1": 1, "User2": 1, "User3": 1, "User4": 1, "User5": 0},  # 4 users
            "07": {"User1": 1, "User2": 0, "User3": 0, "User4": 1, "User5": 2},  # 4 users
            "08": {"User1": 1, "User2": 0, "User3": 0, "User4": 0, "User5": 3},  # 4 users
            "09": {"User1": 1, "User2": 0, "User3": 0, "User4": 0, "User5": 3},  # 4 users
            "10": {"User1": 1, "User2": 0, "User3": 0, "User4": 0, "User5": 3},  # 4 users
        }
        self.assertDictEqual(users_dispatcher._desired_users_assigned_to_workers, expected_balanced_users)


class TestDispatchUsersWithWorkersWithoutPriorUsers(unittest.TestCase):
    def test_dispatch_users_to_3_workers_with_spawn_rate_of_0_5(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=0.5,
        )

        sleep_time = 1 / 0.5

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 0},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_1(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=1,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 0},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_2(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=2,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_2_4(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=2.4,
        )

        sleep_time = 2 / 2.4

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_3(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=3,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 0, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_4(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=4,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 0},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 0, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_9(self):
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=9,
        )

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)


class TestDispatchUsersToWorkersHavingLessUsersThanTheTarget(unittest.TestCase):
    def test_dispatch_users_to_3_workers_with_spawn_rate_of_0_5(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 1}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=0.5,
        )

        sleep_time = 1 / 0.5

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 0, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_1(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 1}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=1,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 0, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_2(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 1}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=2,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_2_4(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 1}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=2.4,
        )

        sleep_time = 2 / 2.4

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_3(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 1}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=3,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_4(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 1}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=4,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 0, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_9(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 1}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=9,
        )

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_5_workers_with_spawn_rate_of_3(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {"User1": 1, "User2": 1, "User3": 0}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1, "User2": 1, "User3": 0}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User1": 1, "User2": 1, "User3": 0}
        worker_node4 = WorkerNode("4")
        worker_node4.user_classes_count = {"User1": 1, "User2": 0, "User3": 1}
        worker_node5 = WorkerNode("5")
        worker_node5.user_classes_count = {"User1": 0, "User2": 0, "User3": 2}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3, worker_node4, worker_node5],
            user_classes_count={"User1": 5, "User2": 5, "User3": 5},
            spawn_rate=3,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 0},
                "3": {"User1": 1, "User2": 1, "User3": 0},
                "4": {"User1": 1, "User2": 1, "User3": 1},
                "5": {"User1": 1, "User2": 0, "User3": 2},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
                "4": {"User1": 1, "User2": 1, "User3": 1},
                "5": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)


class TestDispatchUsersToWorkersHavingLessAndMoreUsersThanTheTargetAndMoreTotalUsers(unittest.TestCase):
    """
    Test scenario with:
        - Already more total users running than desired
        - Some workers already have more users than desired
        - Some workers have less users than desired
    """

    def test_dispatch_users_to_3_workers(self):
        for worker_node1_user_classes_count in [{}, {"User3": 1}]:
            worker_node1 = WorkerNode("1")
            worker_node1.user_classes_count = worker_node1_user_classes_count
            worker_node2 = WorkerNode("2")
            worker_node2.user_classes_count = {"User1": 5}
            worker_node3 = WorkerNode("3")
            worker_node3.user_classes_count = {"User2": 7}

            for spawn_rate in [0.5, 1, 2, 2.4, 3, 4, 9]:
                users_dispatcher = UsersDispatcher(
                    worker_nodes=[worker_node1, worker_node2, worker_node3],
                    user_classes_count={"User1": 3, "User2": 3, "User3": 3},
                    spawn_rate=spawn_rate,
                )

                ts = time.perf_counter()
                self.assertDictEqual(
                    next(users_dispatcher),
                    {
                        "1": {"User1": 1, "User2": 1, "User3": 1},
                        "2": {"User1": 1, "User2": 1, "User3": 1},
                        "3": {"User1": 1, "User2": 1, "User3": 1},
                    },
                )
                delta = time.perf_counter() - ts
                self.assertTrue(0 <= delta <= _TOLERANCE, delta)

                ts = time.perf_counter()
                self.assertRaises(StopIteration, lambda: next(users_dispatcher))
                delta = time.perf_counter() - ts
                self.assertTrue(0 <= delta <= _TOLERANCE, delta)

                clear_all_functools_lru_cache()


class TestDispatchUsersToWorkersHavingLessAndMoreUsersThanTheTargetAndLessTotalUsers(unittest.TestCase):
    """
    Test scenario with:
        - Some users are already but there are less total users running than desired
        - Some workers already have more users than desired
        - Some workers have less users than desired
    """

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_0_5_and_one_worker_empty(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 4}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 4}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=0.5,
        )

        sleep_time = 1 / 0.5

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 0, "User2": 0, "User3": 1},
                "2": {"User1": 4, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 4, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_1_and_one_worker_empty(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 4}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 4}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=1,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 0, "User2": 0, "User3": 1},
                "2": {"User1": 4, "User2": 0, "User3": 0},
                "3": {"User1": 0, "User2": 4, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_0_5_and_all_workers_non_empty(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {"User3": 1}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 3}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 4}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=0.5,
        )

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_1_and_all_workers_non_empty(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {"User3": 1}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 3}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 4}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=1,
        )

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_0_5_and_all_workers_non_empty_and_one_user_class_requiring_no_change(
        self,
    ):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {"User1": 1, "User3": 1}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User1": 1, "User2": 4}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=0.5,
        )

        sleep_time = 1 / 0.5

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 1},
                "3": {"User1": 1, "User2": 4, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_of_1_and_all_workers_non_empty_and_one_user_class_requiring_no_change(
        self,
    ):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {"User1": 1, "User3": 1}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User1": 1, "User2": 4}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=1,
        )

        sleep_time = 1

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 0, "User3": 1},
                "2": {"User1": 1, "User2": 0, "User3": 1},
                "3": {"User1": 1, "User2": 4, "User3": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertDictEqual(
            next(users_dispatcher),
            {
                "1": {"User1": 1, "User2": 1, "User3": 1},
                "2": {"User1": 1, "User2": 1, "User3": 1},
                "3": {"User1": 1, "User2": 1, "User3": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_users_to_3_workers_with_spawn_rate_greater_than_or_equal_to_2(self):
        worker_node1_user_classes_count_cases = [{}, {"User3": 1}, {"User3": 2}]
        worker_node2_user_classes_count_cases = [{"User1": 4}, {"User1": 3}, {"User1": 3}]
        worker_node3_user_classes_count_cases = [{"User2": 4}, {"User2": 3}, {"User2": 3}]
        for (worker_node1_user_classes_count, worker_node2_user_classes_count, worker_node3_user_classes_count) in zip(
            worker_node1_user_classes_count_cases,
            worker_node2_user_classes_count_cases,
            worker_node3_user_classes_count_cases,
        ):
            worker_node1 = WorkerNode("1")
            worker_node1.user_classes_count = worker_node1_user_classes_count
            worker_node2 = WorkerNode("2")
            worker_node2.user_classes_count = worker_node2_user_classes_count
            worker_node3 = WorkerNode("3")
            worker_node3.user_classes_count = worker_node3_user_classes_count

            for spawn_rate in [2, 2.4, 3, 4, 9]:
                users_dispatcher = UsersDispatcher(
                    worker_nodes=[worker_node1, worker_node2, worker_node3],
                    user_classes_count={"User1": 3, "User2": 3, "User3": 3},
                    spawn_rate=spawn_rate,
                )

                ts = time.perf_counter()
                self.assertDictEqual(
                    next(users_dispatcher),
                    {
                        "1": {"User1": 1, "User2": 1, "User3": 1},
                        "2": {"User1": 1, "User2": 1, "User3": 1},
                        "3": {"User1": 1, "User2": 1, "User3": 1},
                    },
                )
                delta = time.perf_counter() - ts
                self.assertTrue(0 <= delta <= _TOLERANCE, delta)

                ts = time.perf_counter()
                self.assertRaises(StopIteration, lambda: next(users_dispatcher))
                delta = time.perf_counter() - ts
                self.assertTrue(0 <= delta <= _TOLERANCE, delta)

                clear_all_functools_lru_cache()


class TestDispatchUsersToWorkersHavingMoreUsersThanTheTarget(unittest.TestCase):
    def test_dispatch_users_to_3_workers(self):
        worker_node1_user_classes_count_cases = [{"User3": 15}, {"User3": 3}]
        worker_node2_user_classes_count_cases = [{"User1": 5}, {"User1": 3}]
        worker_node3_user_classes_count_cases = [{"User2": 7}, {"User2": 3}]
        for (
            worker_node1_user_classes_count,
            worker_node2_user_classes_count,
            worker_node3_user_classes_count,
        ) in itertools.product(
            worker_node1_user_classes_count_cases,
            worker_node2_user_classes_count_cases,
            worker_node3_user_classes_count_cases,
        ):
            worker_node1 = WorkerNode("1")
            worker_node1.user_classes_count = worker_node1_user_classes_count
            worker_node2 = WorkerNode("2")
            worker_node2.user_classes_count = worker_node2_user_classes_count
            worker_node3 = WorkerNode("3")
            worker_node3.user_classes_count = worker_node3_user_classes_count

            for spawn_rate in [0.15, 0.5, 1, 2, 2.4, 3, 4, 9]:
                users_dispatcher = UsersDispatcher(
                    worker_nodes=[worker_node1, worker_node2, worker_node3],
                    user_classes_count={"User1": 3, "User2": 3, "User3": 3},
                    spawn_rate=spawn_rate,
                )

                ts = time.perf_counter()
                self.assertDictEqual(
                    next(users_dispatcher),
                    {
                        "1": {"User1": 1, "User2": 1, "User3": 1},
                        "2": {"User1": 1, "User2": 1, "User3": 1},
                        "3": {"User1": 1, "User2": 1, "User3": 1},
                    },
                )
                delta = time.perf_counter() - ts
                self.assertTrue(0 <= delta <= _TOLERANCE, delta)

                ts = time.perf_counter()
                self.assertRaises(StopIteration, lambda: next(users_dispatcher))
                delta = time.perf_counter() - ts
                self.assertTrue(0 <= delta <= _TOLERANCE, delta)

                clear_all_functools_lru_cache()


class TestDispatchUsersToWorkersHavingTheSameUsersAsTheTarget(unittest.TestCase):
    def test_dispatch_users_to_3_workers(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {"User1": 1, "User2": 1, "User3": 1}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {"User1": 1, "User2": 1, "User3": 1}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User1": 1, "User2": 1, "User3": 1}

        for spawn_rate in [0.15, 0.5, 1, 2, 2.4, 3, 4, 9]:
            users_dispatcher = UsersDispatcher(
                worker_nodes=[worker_node1, worker_node2, worker_node3],
                user_classes_count={"User1": 3, "User2": 3, "User3": 3},
                spawn_rate=spawn_rate,
            )

            ts = time.perf_counter()
            self.assertDictEqual(
                next(users_dispatcher),
                {
                    "1": {"User1": 1, "User2": 1, "User3": 1},
                    "2": {"User1": 1, "User2": 1, "User3": 1},
                    "3": {"User1": 1, "User2": 1, "User3": 1},
                },
            )
            delta = time.perf_counter() - ts
            self.assertTrue(0 <= delta <= _TOLERANCE, delta)

            ts = time.perf_counter()
            self.assertRaises(StopIteration, lambda: next(users_dispatcher))
            delta = time.perf_counter() - ts
            self.assertTrue(0 <= delta <= _TOLERANCE, delta)

            clear_all_functools_lru_cache()


class TestDistributionIsKeptDuringDispatch(unittest.TestCase):
    def test_dispatch_75_users_to_4_workers_with_spawn_rate_of_5(self):
        """
        Test case covering reported issue in https://github.com/locustio/locust/pull/1621#issuecomment-853624275.

        The case is to ramp-up from 0 to 75 users with two user classes. `User1` has a weight of 1 and `User2`
        has a weight of 2. The original issue was with 500 users, but to keep the test shorter, we use 75 users.
        """
        worker_node1 = WorkerNode("1")
        worker_node2 = WorkerNode("2")
        worker_node3 = WorkerNode("3")
        worker_node4 = WorkerNode("4")

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3, worker_node4],
            user_classes_count={"User1": 25, "User2": 50},
            spawn_rate=5,
        )

        sleep_time = 1

        # total user count = 5
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 2, "User2": 3})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 1, "User2": 1},
                "2": {"User1": 0, "User2": 1},
                "3": {"User1": 0, "User2": 1},
                "4": {"User1": 1, "User2": 0},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

        # total user count = 10
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 4, "User2": 6})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 1, "User2": 2},
                "2": {"User1": 1, "User2": 2},
                "3": {"User1": 1, "User2": 1},
                "4": {"User1": 1, "User2": 1},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 15
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 5, "User2": 10})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 2, "User2": 2},
                "2": {"User1": 1, "User2": 3},
                "3": {"User1": 1, "User2": 3},
                "4": {"User1": 1, "User2": 2},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 20
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 7, "User2": 13})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 2, "User2": 3},
                "2": {"User1": 1, "User2": 4},
                "3": {"User1": 2, "User2": 3},
                "4": {"User1": 2, "User2": 3},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 25
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 9, "User2": 16})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 3, "User2": 4},
                "2": {"User1": 2, "User2": 4},
                "3": {"User1": 2, "User2": 4},
                "4": {"User1": 2, "User2": 4},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 30
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 10, "User2": 20})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 3, "User2": 5},
                "2": {"User1": 2, "User2": 6},
                "3": {"User1": 2, "User2": 5},
                "4": {"User1": 3, "User2": 4},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 35
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 12, "User2": 23})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 3, "User2": 6},
                "2": {"User1": 3, "User2": 6},
                "3": {"User1": 3, "User2": 6},
                "4": {"User1": 3, "User2": 5},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 40
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 14, "User2": 26})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 4, "User2": 6},
                "2": {"User1": 3, "User2": 7},
                "3": {"User1": 3, "User2": 7},
                "4": {"User1": 4, "User2": 6},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 45
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 15, "User2": 30})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 4, "User2": 8},
                "2": {"User1": 3, "User2": 8},
                "3": {"User1": 4, "User2": 7},
                "4": {"User1": 4, "User2": 7},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 50
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 17, "User2": 33})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 5, "User2": 8},
                "2": {"User1": 4, "User2": 9},
                "3": {"User1": 4, "User2": 8},
                "4": {"User1": 4, "User2": 8},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 55
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 19, "User2": 36})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 5, "User2": 9},
                "2": {"User1": 4, "User2": 10},
                "3": {"User1": 5, "User2": 9},
                "4": {"User1": 5, "User2": 8},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 60
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 20, "User2": 40})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 5, "User2": 10},
                "2": {"User1": 5, "User2": 10},
                "3": {"User1": 5, "User2": 10},
                "4": {"User1": 5, "User2": 10},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 65
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 22, "User2": 43})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 6, "User2": 11},
                "2": {"User1": 5, "User2": 11},
                "3": {"User1": 5, "User2": 11},
                "4": {"User1": 6, "User2": 10},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 70
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 24, "User2": 46})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 6, "User2": 12},
                "2": {"User1": 6, "User2": 12},
                "3": {"User1": 6, "User2": 11},
                "4": {"User1": 6, "User2": 11},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        # total user count = 75, User1 = 25, User2 = 50
        ts = time.perf_counter()
        dispatched_users = next(users_dispatcher)
        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 25, "User2": 50})
        self.assertDictEqual(
            dispatched_users,
            {
                "1": {"User1": 7, "User2": 12},
                "2": {"User1": 6, "User2": 13},
                "3": {"User1": 6, "User2": 13},
                "4": {"User1": 6, "User2": 12},
            },
        )
        delta = time.perf_counter() - ts
        self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)


class TestDispatch(unittest.TestCase):
    """
    Class containing miscellaneous tests that were used to debug some issues observed when running real tests.
    """

    def test_dispatch_50_total_users_with_25_already_running_to_20_workers_with_spawn_rate_of_1(self):
        """
        Prior to 81f6d66cf69755ee966a26741e251ac560c78233, the dispatcher would have directly go from 25 to 50 users
        in one second due to bugs in the dispatcher. This test ensures that this problem can't
        reappear in the future.
        """
        # Prepend "0" to worker name under 10, because workers are sorted in alphabetical orders.
        # It's simply makes the test easier to read and debug, but in reality, the worker "10" would come
        # before worker "1".
        worker_nodes = [WorkerNode("0{}".format(i) if i < 10 else str(i)) for i in range(1, 21)]

        for worker_node in worker_nodes:
            worker_node.user_classes_count = {"User1": 0}

        worker_nodes_iterator = itertools.cycle(worker_nodes)
        user_count = 0
        while user_count < 25:
            next(worker_nodes_iterator).user_classes_count["User1"] += 1
            user_count += 1

        users_dispatcher = UsersDispatcher(worker_nodes=worker_nodes, user_classes_count={"User1": 50}, spawn_rate=1)

        sleep_time = 1

        for dispatch_iteration in range(25):
            ts = time.perf_counter()
            dispatched_users = next(users_dispatcher)
            self.assertDictEqual(_aggregate_dispatched_users(dispatched_users), {"User1": 25 + dispatch_iteration + 1})
            delta = time.perf_counter() - ts
            if dispatch_iteration == 0:
                self.assertTrue(0 <= delta <= _TOLERANCE, delta)
            else:
                self.assertTrue(sleep_time - _TOLERANCE <= delta <= sleep_time + _TOLERANCE, delta)

        ts = time.perf_counter()
        self.assertRaises(StopIteration, lambda: next(users_dispatcher))
        delta = time.perf_counter() - ts
        self.assertTrue(0 <= delta <= _TOLERANCE, delta)

    def test_dispatch_from_5_to_10_users_to_10_workers(self):
        """
        This test case was causing the dispatcher to go into an infinite loop.
        The test is left to prevent future regressions.
        """
        worker_nodes = [WorkerNode("0{}".format(i) if i < 10 else str(i)) for i in range(1, 11)]
        worker_nodes[1].user_classes_count = {"TestUser6": 1}
        worker_nodes[3].user_classes_count = {"TestUser4": 1}
        worker_nodes[4].user_classes_count = {"TestUser5": 1}
        worker_nodes[5].user_classes_count = {"TestUser10": 1}
        worker_nodes[9].user_classes_count = {"TestUser1": 1}

        user_classes_count = {
            "TestUser1": 1,
            "TestUser2": 1,
            "TestUser3": 1,
            "TestUser4": 1,
            "TestUser5": 1,
            "TestUser6": 1,
            "TestUser7": 1,
            "TestUser8": 1,
            "TestUser9": 1,
            "TestUser10": 1,
        }

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes, user_classes_count=user_classes_count, spawn_rate=1
        )

        dispatched_users = list(users_dispatcher)

        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users[-1]), user_classes_count)

    def test_dispatch_from_20_to_55_users_to_10_workers(self):
        """
        This test case was causing the dispatcher to go into an infinite loop.
        The test is left to prevent future regressions.
        """
        worker_nodes = [WorkerNode("0{}".format(i) if i < 10 else str(i)) for i in range(1, 11)]
        worker_nodes[0].user_classes_count = {
            "TestUser1": 1,
            "TestUser10": 1,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 0,
            "TestUser5": 0,
            "TestUser6": 0,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 0,
        }
        worker_nodes[1].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 1,
            "TestUser2": 1,
            "TestUser3": 0,
            "TestUser4": 0,
            "TestUser5": 0,
            "TestUser6": 0,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 0,
        }
        worker_nodes[2].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 1,
            "TestUser3": 1,
            "TestUser4": 0,
            "TestUser5": 0,
            "TestUser6": 0,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 0,
        }
        worker_nodes[3].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 1,
            "TestUser5": 1,
            "TestUser6": 0,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 0,
        }
        worker_nodes[4].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 1,
            "TestUser5": 1,
            "TestUser6": 0,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 0,
        }
        worker_nodes[5].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 1,
            "TestUser5": 1,
            "TestUser6": 0,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 0,
        }
        worker_nodes[6].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 0,
            "TestUser5": 0,
            "TestUser6": 1,
            "TestUser7": 1,
            "TestUser8": 0,
            "TestUser9": 0,
        }
        worker_nodes[7].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 0,
            "TestUser5": 0,
            "TestUser6": 1,
            "TestUser7": 0,
            "TestUser8": 1,
            "TestUser9": 0,
        }
        worker_nodes[8].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 0,
            "TestUser5": 0,
            "TestUser6": 1,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 1,
        }
        worker_nodes[9].user_classes_count = {
            "TestUser1": 0,
            "TestUser10": 0,
            "TestUser2": 0,
            "TestUser3": 0,
            "TestUser4": 0,
            "TestUser5": 0,
            "TestUser6": 1,
            "TestUser7": 0,
            "TestUser8": 0,
            "TestUser9": 1,
        }

        user_classes_count = {
            "TestUser1": 2,
            "TestUser10": 5,
            "TestUser2": 5,
            "TestUser3": 3,
            "TestUser4": 8,
            "TestUser5": 10,
            "TestUser6": 11,
            "TestUser7": 4,
            "TestUser8": 2,
            "TestUser9": 5,
        }

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes, user_classes_count=user_classes_count, spawn_rate=2
        )

        dispatched_users = list(users_dispatcher)

        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users[-1]), user_classes_count)

    def test_dispatch_from_38_to_61_users_to_10_workers(self):
        """
        This test case was causing the dispatcher to go into an infinite loop.
        The test is left to prevent future regressions.
        """
        worker_nodes = [WorkerNode("0{}".format(i) if i < 10 else str(i)) for i in range(1, 11)]
        worker_nodes[0].user_classes_count = {
            "TestUser01": 1,
            "TestUser02": 1,
            "TestUser03": 1,
            "TestUser05": 0,
            "TestUser06": 0,
            "TestUser07": 0,
            "TestUser08": 0,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 0,
            "TestUser12": 1,
            "TestUser14": 0,
            "TestUser15": 0,
        }
        worker_nodes[1].user_classes_count = {
            "TestUser01": 1,
            "TestUser02": 1,
            "TestUser03": 1,
            "TestUser05": 0,
            "TestUser06": 0,
            "TestUser07": 0,
            "TestUser08": 0,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 0,
            "TestUser12": 0,
            "TestUser14": 1,
            "TestUser15": 0,
        }
        worker_nodes[2].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 1,
            "TestUser03": 1,
            "TestUser05": 1,
            "TestUser06": 0,
            "TestUser07": 0,
            "TestUser08": 0,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 0,
            "TestUser12": 0,
            "TestUser14": 1,
            "TestUser15": 0,
        }
        worker_nodes[3].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 0,
            "TestUser03": 1,
            "TestUser05": 1,
            "TestUser06": 1,
            "TestUser07": 0,
            "TestUser08": 0,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 0,
            "TestUser12": 0,
            "TestUser14": 0,
            "TestUser15": 1,
        }
        worker_nodes[4].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 0,
            "TestUser03": 0,
            "TestUser05": 1,
            "TestUser06": 1,
            "TestUser07": 1,
            "TestUser08": 0,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 0,
            "TestUser12": 0,
            "TestUser14": 0,
            "TestUser15": 1,
        }
        worker_nodes[5].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 0,
            "TestUser03": 0,
            "TestUser05": 1,
            "TestUser06": 0,
            "TestUser07": 1,
            "TestUser08": 1,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 0,
            "TestUser12": 0,
            "TestUser14": 0,
            "TestUser15": 1,
        }
        worker_nodes[6].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 0,
            "TestUser03": 0,
            "TestUser05": 0,
            "TestUser06": 0,
            "TestUser07": 1,
            "TestUser08": 1,
            "TestUser09": 1,
            "TestUser10": 0,
            "TestUser11": 0,
            "TestUser12": 0,
            "TestUser14": 0,
            "TestUser15": 1,
        }
        worker_nodes[7].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 0,
            "TestUser03": 0,
            "TestUser05": 0,
            "TestUser06": 0,
            "TestUser07": 0,
            "TestUser08": 1,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 1,
            "TestUser12": 1,
            "TestUser14": 0,
            "TestUser15": 1,
        }
        worker_nodes[8].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 0,
            "TestUser03": 0,
            "TestUser05": 0,
            "TestUser06": 0,
            "TestUser07": 0,
            "TestUser08": 1,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 1,
            "TestUser12": 1,
            "TestUser14": 0,
            "TestUser15": 0,
        }
        worker_nodes[9].user_classes_count = {
            "TestUser01": 0,
            "TestUser02": 0,
            "TestUser03": 0,
            "TestUser05": 0,
            "TestUser06": 0,
            "TestUser07": 0,
            "TestUser08": 0,
            "TestUser09": 0,
            "TestUser10": 0,
            "TestUser11": 1,
            "TestUser12": 2,
            "TestUser14": 0,
            "TestUser15": 0,
        }

        user_classes_count = {
            "TestUser01": 3,
            "TestUser02": 4,
            "TestUser03": 7,
            "TestUser05": 7,
            "TestUser06": 3,
            "TestUser07": 6,
            "TestUser08": 6,
            "TestUser09": 1,
            "TestUser10": 0,
            "TestUser11": 5,
            "TestUser12": 7,
            "TestUser14": 4,
            "TestUser15": 8,
        }

        users_dispatcher = UsersDispatcher(
            worker_nodes=worker_nodes, user_classes_count=user_classes_count, spawn_rate=8.556688078766006
        )

        dispatched_users = list(users_dispatcher)

        self.assertDictEqual(_aggregate_dispatched_users(dispatched_users[-1]), user_classes_count)

    def test_crash_dump(self):
        worker_node1 = WorkerNode("1")
        worker_node1.user_classes_count = {"User3": 1}
        worker_node2 = WorkerNode("2")
        worker_node2.user_classes_count = {}
        worker_node3 = WorkerNode("3")
        worker_node3.user_classes_count = {"User2": 4}

        users_dispatcher = UsersDispatcher(
            worker_nodes=[worker_node1, worker_node2, worker_node3],
            user_classes_count={"User1": 3, "User2": 3, "User3": 3},
            spawn_rate=0.7,
        )

        crash_dump_content, crash_dump_filepath = users_dispatcher._crash_dump()

        self.assertIsInstance(crash_dump_content, str)
        self.assertIsInstance(crash_dump_filepath, str)

        crash_dumps = [json.loads(crash_dump_content)]

        with open(crash_dump_filepath, "rt") as f:
            crash_dumps.append(json.load(f))

        for crash_dump in crash_dumps:
            self.assertEqual(len(crash_dump), 7)
            self.assertEqual(crash_dump["spawn_rate"], 0.7)
            self.assertDictEqual(
                crash_dump["initial_dispatched_users"],
                {
                    "1": {"User1": 0, "User2": 0, "User3": 1},
                    "2": {"User1": 0, "User2": 0, "User3": 0},
                    "3": {"User1": 0, "User2": 4, "User3": 0},
                },
            )
            self.assertDictEqual(
                crash_dump["desired_users_assigned_to_workers"],
                {
                    "1": {"User1": 1, "User2": 1, "User3": 1},
                    "2": {"User1": 1, "User2": 1, "User3": 1},
                    "3": {"User1": 1, "User2": 1, "User3": 1},
                },
            )
            self.assertDictEqual(crash_dump["user_classes_count"], {"User1": 3, "User2": 3, "User3": 3})
            self.assertEqual(crash_dump["initial_user_count"], 5)
            self.assertEqual(crash_dump["desired_user_count"], 9)
            self.assertEqual(crash_dump["number_of_workers"], 3)


def _aggregate_dispatched_users(d: Dict[str, Dict[str, int]]) -> Dict[str, int]:
    user_classes = list(next(iter(d.values())).keys())
    return {u: sum(d[u] for d in d.values()) for u in user_classes}



__version__ = "2.0.0b1"

from gevent import monkey

monkey.patch_all()

from .user.sequential_taskset import SequentialTaskSet
from .user import wait_time
from .user.task import task, tag, TaskSet
from .user.users import HttpUser, User
from .user.wait_time import between, constant, constant_pacing
from .shape import LoadTestShape

from .event import Events

events = Events()

__all__ = (
    "SequentialTaskSet",
    "wait_time",
    "task",
    "tag",
    "TaskSet",
    "HttpUser",
    "User",
    "between",
    "constant",
    "constant_pacing",
    "events",
    "LoadTestShape",
)

# Used for raising a DeprecationWarning if old Locust/HttpLocust is used
from .util.deprecation import DeprecatedLocustClass as Locust
from .util.deprecation import DeprecatedHttpLocustClass as HttpLocust



import argparse
import os
import sys
import textwrap

import configargparse

import locust

version = locust.__version__


DEFAULT_CONFIG_FILES = ["~/.locust.conf", "locust.conf"]


def _is_package(path):
    """
    Is the given path a Python package?
    """
    return os.path.isdir(path) and os.path.exists(os.path.join(path, "__init__.py"))


def find_locustfile(locustfile):
    """
    Attempt to locate a locustfile, either explicitly or by searching parent dirs.
    """
    # Obtain env value
    names = [locustfile]
    # Create .py version if necessary
    if not names[0].endswith(".py"):
        names.append(names[0] + ".py")
    # Does the name contain path elements?
    if os.path.dirname(names[0]):
        # If so, expand home-directory markers and test for existence
        for name in names:
            expanded = os.path.expanduser(name)
            if os.path.exists(expanded):
                if name.endswith(".py") or _is_package(expanded):
                    return os.path.abspath(expanded)
    else:
        # Otherwise, start in cwd and work downwards towards filesystem root
        path = os.path.abspath(".")
        while True:
            for name in names:
                joined = os.path.join(path, name)
                if os.path.exists(joined):
                    if name.endswith(".py") or _is_package(joined):
                        return os.path.abspath(joined)
            parent_path = os.path.dirname(path)
            if parent_path == path:
                # we've reached the root path which has been checked this iteration
                break
            path = parent_path
    # Implicit 'return None' if nothing was found


def get_empty_argument_parser(add_help=True, default_config_files=DEFAULT_CONFIG_FILES):
    parser = configargparse.ArgumentParser(
        default_config_files=default_config_files,
        add_env_var_help=False,
        add_config_file_help=False,
        add_help=add_help,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        usage=argparse.SUPPRESS,
        description=textwrap.dedent(
            """
            Usage: locust [OPTIONS] [UserClass ...]

        """
        ),
        # epilog="",
    )
    parser.add_argument(
        "-f",
        "--locustfile",
        default="locustfile",
        help="Python module file to import, e.g. '../other.py'. Default: locustfile",
        env_var="LOCUST_LOCUSTFILE",
    )
    parser.add_argument("--config", is_config_file_arg=True, help="Config file path")

    return parser


def parse_locustfile_option(args=None):
    """
    Construct a command line parser that is only used to parse the -f argument so that we can
    import the test scripts in case any of them adds additional command line arguments to the
    parser
    """
    parser = get_empty_argument_parser(add_help=False)
    parser.add_argument(
        "-h",
        "--help",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--version",
        "-V",
        action="store_true",
        default=False,
    )

    options, _ = parser.parse_known_args(args=args)

    locustfile = find_locustfile(options.locustfile)

    if not locustfile:
        if options.help or options.version:
            # if --help or --version is specified we'll call parse_options which will print the help/version message
            parse_options(args=args)
        sys.stderr.write(
            "Could not find any locustfile! Ensure file ends in '.py' and see --help for available options.\n"
        )
        sys.exit(1)

    if locustfile == "locust.py":
        sys.stderr.write("The locustfile must not be named `locust.py`. Please rename the file and try again.\n")
        sys.exit(1)

    return locustfile


def setup_parser_arguments(parser):
    """
    Setup command-line options

    Takes a configargparse.ArgumentParser as argument and calls it's add_argument
    for each of the supported arguments
    """
    parser._optionals.title = "Common options"
    parser.add_argument(
        "-H",
        "--host",
        help="Host to load test in the following format: http://10.21.32.33",
        env_var="LOCUST_HOST",
    )
    parser.add_argument(
        "-u",
        "--users",
        type=int,
        dest="num_users",
        help="Number of concurrent Locust users. Primarily used together with --headless. Can be changed during a test by inputs w, W(spawn 1, 10 users) and s, S(stop 1, 10 users)",
        env_var="LOCUST_USERS",
    )
    parser.add_argument(
        "-r",
        "--spawn-rate",
        type=float,
        help="The rate per second in which users are spawned. Primarily used together with --headless",
        env_var="LOCUST_SPAWN_RATE",
    )
    parser.add_argument(
        "--hatch-rate",
        env_var="LOCUST_HATCH_RATE",
        type=float,
        default=0,
        help=configargparse.SUPPRESS,
    )
    parser.add_argument(
        "-t",
        "--run-time",
        help="Stop after the specified amount of time, e.g. (300s, 20m, 3h, 1h30m, etc.). Only used together with --headless. Defaults to run forever.",
        env_var="LOCUST_RUN_TIME",
    )
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        dest="list_commands",
        help="Show list of possible User classes and exit",
    )

    web_ui_group = parser.add_argument_group("Web UI options")
    web_ui_group.add_argument(
        "--web-host",
        default="",
        help="Host to bind the web interface to. Defaults to '*' (all interfaces)",
        env_var="LOCUST_WEB_HOST",
    )
    web_ui_group.add_argument(
        "--web-port",
        "-P",
        type=int,
        default=8089,
        help="Port on which to run web host",
        env_var="LOCUST_WEB_PORT",
    )
    web_ui_group.add_argument(
        "--headless",
        action="store_true",
        help="Disable the web interface, and instead start the load test immediately. Requires -u and -t to be specified.",
        env_var="LOCUST_HEADLESS",
    )
    # Override --headless parameter (useful because you cant disable a store_true-parameter like headless once it has been set in a config file)
    web_ui_group.add_argument(
        "--headful",
        action="store_true",
        help=configargparse.SUPPRESS,
        env_var="LOCUST_HEADFUL",
    )
    web_ui_group.add_argument(
        "--web-auth",
        type=str,
        dest="web_auth",
        default=None,
        help="Turn on Basic Auth for the web interface. Should be supplied in the following format: username:password",
        env_var="LOCUST_WEB_AUTH",
    )
    web_ui_group.add_argument(
        "--tls-cert",
        default="",
        help="Optional path to TLS certificate to use to serve over HTTPS",
        env_var="LOCUST_TLS_CERT",
    )
    web_ui_group.add_argument(
        "--tls-key",
        default="",
        help="Optional path to TLS private key to use to serve over HTTPS",
        env_var="LOCUST_TLS_KEY",
    )

    master_group = parser.add_argument_group(
        "Master options",
        "Options for running a Locust Master node when running Locust distributed. A Master node need Worker nodes that connect to it before it can run load tests.",
    )
    # if locust should be run in distributed mode as master
    master_group.add_argument(
        "--master",
        action="store_true",
        help="Set locust to run in distributed mode with this process as master",
        env_var="LOCUST_MODE_MASTER",
    )
    master_group.add_argument(
        "--master-bind-host",
        default="*",
        help="Interfaces (hostname, ip) that locust master should bind to. Only used when running with --master. Defaults to * (all available interfaces).",
        env_var="LOCUST_MASTER_BIND_HOST",
    )
    master_group.add_argument(
        "--master-bind-port",
        type=int,
        default=5557,
        help="Port that locust master should bind to. Only used when running with --master. Defaults to 5557.",
        env_var="LOCUST_MASTER_BIND_PORT",
    )
    master_group.add_argument(
        "--expect-workers",
        type=int,
        default=1,
        help="How many workers master should expect to connect before starting the test (only when --headless used).",
        env_var="LOCUST_EXPECT_WORKERS",
    )
    master_group.add_argument(
        "--expect-slaves",
        action="store_true",
        help=configargparse.SUPPRESS,
    )

    worker_group = parser.add_argument_group(
        "Worker options",
        textwrap.dedent(
            """
            Options for running a Locust Worker node when running Locust distributed.
            Only the LOCUSTFILE (-f option) need to be specified when starting a Worker, since other options such as -u, -r, -t are specified on the Master node.
        """
        ),
    )
    # if locust should be run in distributed mode as worker
    worker_group.add_argument(
        "--worker",
        action="store_true",
        help="Set locust to run in distributed mode with this process as worker",
        env_var="LOCUST_MODE_WORKER",
    )
    worker_group.add_argument(
        "--slave",
        action="store_true",
        help=configargparse.SUPPRESS,
    )
    # master host options
    worker_group.add_argument(
        "--master-host",
        default="127.0.0.1",
        help="Host or IP address of locust master for distributed load testing. Only used when running with --worker. Defaults to 127.0.0.1.",
        env_var="LOCUST_MASTER_NODE_HOST",
        metavar="MASTER_NODE_HOST",
    )
    worker_group.add_argument(
        "--master-port",
        type=int,
        default=5557,
        help="The port to connect to that is used by the locust master for distributed load testing. Only used when running with --worker. Defaults to 5557.",
        env_var="LOCUST_MASTER_NODE_PORT",
        metavar="MASTER_NODE_PORT",
    )

    tag_group = parser.add_argument_group(
        "Tag options",
        "Locust tasks can be tagged using the @tag decorator. These options let specify which tasks to include or exclude during a test.",
    )
    tag_group.add_argument(
        "-T",
        "--tags",
        nargs="*",
        metavar="TAG",
        env_var="LOCUST_TAGS",
        help="List of tags to include in the test, so only tasks with any matching tags will be executed",
    )
    tag_group.add_argument(
        "-E",
        "--exclude-tags",
        nargs="*",
        metavar="TAG",
        env_var="LOCUST_EXCLUDE_TAGS",
        help="List of tags to exclude from the test, so only tasks with no matching tags will be executed",
    )

    stats_group = parser.add_argument_group("Request statistics options")
    stats_group.add_argument(
        "--csv",  # Name repeated in 'parse_options'
        dest="csv_prefix",
        help="Store current request stats to files in CSV format. Setting this option will generate three files: [CSV_PREFIX]_stats.csv, [CSV_PREFIX]_stats_history.csv and [CSV_PREFIX]_failures.csv",
        env_var="LOCUST_CSV",
    )
    stats_group.add_argument(
        "--csv-full-history",  # Name repeated in 'parse_options'
        action="store_true",
        default=False,
        dest="stats_history_enabled",
        help="Store each stats entry in CSV format to _stats_history.csv file. You must also specify the '--csv' argument to enable this.",
        env_var="LOCUST_CSV_FULL_HISTORY",
    )
    stats_group.add_argument(
        "--print-stats",
        action="store_true",
        help="Print stats in the console",
        env_var="LOCUST_PRINT_STATS",
    )
    stats_group.add_argument(
        "--only-summary",
        action="store_true",
        help="Only print the summary stats",
        env_var="LOCUST_ONLY_SUMMARY",
    )
    stats_group.add_argument(
        "--reset-stats",
        action="store_true",
        help="Reset statistics once spawning has been completed. Should be set on both master and workers when running in distributed mode",
        env_var="LOCUST_RESET_STATS",
    )
    stats_group.add_argument(
        "--html",
        dest="html_file",
        help="Store HTML report file",
        env_var="LOCUST_HTML",
    )

    log_group = parser.add_argument_group("Logging options")
    log_group.add_argument(
        "--skip-log-setup",
        action="store_true",
        dest="skip_log_setup",
        default=False,
        help="Disable Locust's logging setup. Instead, the configuration is provided by the Locust test or Python defaults.",
        env_var="LOCUST_SKIP_LOG_SETUP",
    )
    log_group.add_argument(
        "--loglevel",
        "-L",
        default="INFO",
        help="Choose between DEBUG/INFO/WARNING/ERROR/CRITICAL. Default is INFO.",
        env_var="LOCUST_LOGLEVEL",
    )
    log_group.add_argument(
        "--logfile",
        help="Path to log file. If not set, log will go to stdout/stderr",
        env_var="LOCUST_LOGFILE",
    )

    step_load_group = parser.add_argument_group("Step load options")
    step_load_group.add_argument("--step-load", action="store_true", help=configargparse.SUPPRESS)
    step_load_group.add_argument("--step-users", type=int, help=configargparse.SUPPRESS)
    step_load_group.add_argument("--step-clients", action="store_true", help=configargparse.SUPPRESS)
    step_load_group.add_argument("--step-time", help=configargparse.SUPPRESS)

    other_group = parser.add_argument_group("Other options")
    other_group.add_argument(
        "--show-task-ratio", action="store_true", help="Print table of the User classes' task execution ratio"
    )
    other_group.add_argument(
        "--show-task-ratio-json", action="store_true", help="Print json data of the User classes' task execution ratio"
    )
    # optparse gives you --version but we have to do it ourselves to get -V too
    other_group.add_argument(
        "--version",
        "-V",
        action="version",
        help="Show program's version number and exit",
        version="%(prog)s {}".format(version),
    )
    other_group.add_argument(
        "--exit-code-on-error",
        type=int,
        default=1,
        help="Sets the process exit code to use when a test result contain any failure or error",
        env_var="LOCUST_EXIT_CODE_ON_ERROR",
    )
    other_group.add_argument(
        "-s",
        "--stop-timeout",
        action="store",
        type=int,
        dest="stop_timeout",
        default=None,
        help="Number of seconds to wait for a simulated user to complete any executing task before exiting. Default is to terminate immediately. This parameter only needs to be specified for the master process when running Locust distributed.",
        env_var="LOCUST_STOP_TIMEOUT",
    )

    user_classes_group = parser.add_argument_group("User classes")
    user_classes_group.add_argument(
        "user_classes",
        nargs="*",
        metavar="UserClass",
        help="Optionally specify which User classes that should be used (available User classes can be listed with -l or --list)",
    )


def get_parser(default_config_files=DEFAULT_CONFIG_FILES):
    # get a parser that is only able to parse the -f argument
    parser = get_empty_argument_parser(add_help=True, default_config_files=default_config_files)
    # add all the other supported arguments
    setup_parser_arguments(parser)
    # fire event to provide a hook for locustscripts and plugins to add command line arguments
    locust.events.init_command_line_parser.fire(parser=parser)
    return parser


def parse_options(args=None):
    parser = get_parser()
    parsed_opts = parser.parse_args(args=args)
    if parsed_opts.stats_history_enabled and (parsed_opts.csv_prefix is None):
        parser.error("'--csv-full-history' requires '--csv'.")
    return parsed_opts



import math
from itertools import combinations_with_replacement
from operator import attrgetter
from typing import Dict, List, Type

from locust import User


def weight_users(user_classes: List[Type[User]], user_count: int) -> Dict[str, int]:
    """
    Compute the desired state of users using the weight of each user class.

    :param user_classes: the list of user class
    :param user_count: total number of users
    :return: the set of users to run
    """
    assert user_count >= 0

    if len(user_classes) == 0:
        return {}

    user_classes = sorted(user_classes, key=attrgetter("__name__"))

    user_classes_count = {user_class.__name__: 0 for user_class in user_classes}

    # If the number of users is less than the number of user classes, at most one user of each user class
    # is chosen. User classes with higher weight are chosen first.
    if user_count <= len(user_classes):
        user_classes_count.update(
            {
                user_class.__name__: 1
                for user_class in sorted(user_classes, key=attrgetter("weight"), reverse=True)[:user_count]
            }
        )
        return user_classes_count

    # If the number of users is greater than or equal to the number of user classes, at least one user of each
    # user class will be chosen. The greater number of users is, the better the actual distribution
    # of users will match the desired one (as dictated by the weight attributes).
    weights = list(map(attrgetter("weight"), user_classes))
    relative_weights = [weight / sum(weights) for weight in weights]
    user_classes_count = {
        user_class.__name__: round(relative_weight * user_count) or 1
        for user_class, relative_weight in zip(user_classes, relative_weights)
    }

    if sum(user_classes_count.values()) == user_count:
        return user_classes_count

    else:
        user_classes_count = _find_ideal_users_to_add_or_remove(
            user_classes, user_count - sum(user_classes_count.values()), user_classes_count
        )
        assert sum(user_classes_count.values()) == user_count
        return user_classes_count


def _find_ideal_users_to_add_or_remove(
    user_classes: List[Type[User]], user_count_to_add_or_remove: int, user_classes_count: Dict[str, int]
) -> Dict[str, int]:
    sign = -1 if user_count_to_add_or_remove < 0 else 1

    user_count_to_add_or_remove = abs(user_count_to_add_or_remove)

    assert user_count_to_add_or_remove <= len(user_classes), user_count_to_add_or_remove

    # Formula for combination with replacement
    # (https://www.tutorialspoint.com/statistics/combination_with_replacement.htm)
    number_of_combinations = math.factorial(len(user_classes) + user_count_to_add_or_remove - 1) / (
        math.factorial(user_count_to_add_or_remove) * math.factorial(len(user_classes) - 1)
    )

    # If the number of combinations with replacement is above this threshold, we simply add/remove
    # users for the first "number_of_users_to_add_or_remove" users. Otherwise, computing the best
    # distribution is too expensive in terms of computation.
    max_number_of_combinations_threshold = 1000

    if number_of_combinations <= max_number_of_combinations_threshold:
        user_classes_count_candidates: Dict[float, Dict[str, int]] = {}
        for user_classes_combination in combinations_with_replacement(user_classes, user_count_to_add_or_remove):
            # Copy in order to not mutate `user_classes_count` for the parent scope
            user_classes_count_candidate = user_classes_count.copy()
            for user_class in user_classes_combination:
                user_classes_count_candidate[user_class.__name__] += sign
            distance = distance_from_desired_distribution(user_classes, user_classes_count_candidate)
            if distance not in user_classes_count_candidates:
                user_classes_count_candidates[distance] = user_classes_count_candidate.copy()

        return user_classes_count_candidates[min(user_classes_count_candidates.keys())]

    else:
        # Copy in order to not mutate `user_classes_count` for the parent scope
        user_classes_count_candidate = user_classes_count.copy()
        for user_class in user_classes[:user_count_to_add_or_remove]:
            user_classes_count_candidate[user_class.__name__] += sign
        return user_classes_count_candidate


def distance_from_desired_distribution(user_classes: List[Type[User]], user_classes_count: Dict[str, int]) -> float:
    actual_ratio_of_user_class = {
        user_class: user_class_count / sum(user_classes_count.values())
        for user_class, user_class_count in user_classes_count.items()
    }

    expected_ratio_of_user_class = {
        user_class.__name__: user_class.weight / sum(map(attrgetter("weight"), user_classes))
        for user_class in user_classes
    }

    differences = [
        actual_ratio_of_user_class[user_class] - expected_ratio
        for user_class, expected_ratio in expected_ratio_of_user_class.items()
    ]

    return math.sqrt(math.fsum(map(lambda x: x ** 2, differences)))



from locust import (
    constant,
)
from locust.env import Environment
from locust.user import (
    User,
    task,
)
from .testcases import LocustTestCase
from .fake_module1_for_env_test import MyUserWithSameName as MyUserWithSameName1
from .fake_module2_for_env_test import MyUserWithSameName as MyUserWithSameName2


class TestEnvironment(LocustTestCase):
    def test_user_classes_count(self):
        class MyUser1(User):
            wait_time = constant(0)

            @task
            def my_task(self):
                pass

        class MyUser2(User):
            wait_time = constant(0)

            @task
            def my_task(self):
                pass

        environment = Environment(user_classes=[MyUser1, MyUser2])

        self.assertDictEqual({"MyUser1": MyUser1, "MyUser2": MyUser2}, environment.user_classes_by_name)

    def test_user_classes_with_same_name_is_error(self):
        with self.assertRaises(ValueError) as e:
            environment = Environment(user_classes=[MyUserWithSameName1, MyUserWithSameName2])

        self.assertEqual(
            e.exception.args[0],
            "The following user classes have the same class name: locust.test.fake_module1_for_env_test.MyUserWithSameName, locust.test.fake_module2_for_env_test.MyUserWithSameName",
        )



import mock
import socket
import subprocess
import textwrap
from logging import getLogger

import gevent

from locust import log
from locust.log import greenlet_exception_logger
from .testcases import LocustTestCase
from .util import temporary_file


class TestGreenletExceptionLogger(LocustTestCase):
    # Gevent outputs all unhandled exceptions to stderr, so we'll suppress that in this test
    @mock.patch("sys.stderr.write")
    def test_greenlet_exception_logger(self, mocked_stderr):
        self.assertFalse(log.unhandled_greenlet_exception)

        def thread():
            raise ValueError("Boom!?")

        logger = getLogger("greenlet_test_logger")
        g = gevent.spawn(thread)
        g.link_exception(greenlet_exception_logger(logger))
        g.join()
        self.assertEqual(1, len(self.mocked_log.critical))
        msg = self.mocked_log.critical[0]
        self.assertIn("Unhandled exception in greenlet: ", msg["message"])
        self.assertTrue(isinstance(msg["exc_info"][1], ValueError))
        self.assertIn("Boom!?", str(msg["exc_info"][1]))
        self.assertTrue(log.unhandled_greenlet_exception)


class TestLoggingOptions(LocustTestCase):
    def test_logging_output(self):
        with temporary_file(
            textwrap.dedent(
                """
            import logging
            from locust import User, task, constant

            custom_logger = logging.getLogger("custom_logger")

            class MyUser(User):
                wait_time = constant(2)
                @task
                def my_task(self):
                    print("running my_task")
                    logging.info("custom log message")
                    custom_logger.info("test")
        """
            )
        ) as file_path:
            output = subprocess.check_output(
                [
                    "locust",
                    "-f",
                    file_path,
                    "-u",
                    "1",
                    "-r",
                    "1",
                    "-t",
                    "1",
                    "--headless",
                ],
                stderr=subprocess.STDOUT,
                timeout=10,
            ).decode("utf-8")

        self.assertIn(
            "%s/INFO/locust.main: Run time limit set to 1 seconds" % socket.gethostname(),
            output,
        )
        self.assertIn(
            "%s/INFO/locust.main: Time limit reached. Stopping Locust." % socket.gethostname(),
            output,
        )
        self.assertIn(
            "%s/INFO/locust.main: Shutting down (exit code 0), bye." % socket.gethostname(),
            output,
        )
        self.assertIn(
            "\nrunning my_task\n",
            output,
        )
        # check that custom message of root logger is also printed
        self.assertIn(
            "%s/INFO/root: custom log message" % socket.gethostname(),
            output,
        )
        # check that custom message of custom_logger is also printed
        self.assertIn(
            "%s/INFO/custom_logger: test" % socket.gethostname(),
            output,
        )

    def test_skip_logging(self):
        with temporary_file(
            textwrap.dedent(
                """
            from locust import User, task, constant

            class MyUser(User):
                wait_time = constant(2)
                @task
                def my_task(self):
                    print("running my_task")
        """
            )
        ) as file_path:
            output = subprocess.check_output(
                [
                    "locust",
                    "-f",
                    file_path,
                    "-u",
                    "1",
                    "-r",
                    "1",
                    "-t",
                    "1",
                    "--headless",
                    "--skip-log-setup",
                ],
                stderr=subprocess.STDOUT,
                timeout=10,
            ).decode("utf-8")
        self.assertEqual("running my_task", output.strip())

    def test_log_to_file(self):
        with temporary_file(
            textwrap.dedent(
                """
            import logging
            from locust import User, task, constant

            class MyUser(User):
                wait_time = constant(2)
                @task
                def my_task(self):
                    print("running my_task")
                    logging.info("custom log message")
        """
            )
        ) as file_path:
            with temporary_file("", suffix=".log") as log_file_path:
                try:
                    output = subprocess.check_output(
                        [
                            "locust",
                            "-f",
                            file_path,
                            "-u",
                            "1",
                            "-r",
                            "1",
                            "-t",
                            "1",
                            "--headless",
                            "--logfile",
                            log_file_path,
                        ],
                        stderr=subprocess.STDOUT,
                        timeout=10,
                    ).decode("utf-8")
                except subprocess.CalledProcessError as e:
                    raise AssertionError(
                        "Running locust command failed. Output was:\n\n%s" % e.stdout.decode("utf-8")
                    ) from e

                with open(log_file_path, encoding="utf-8") as f:
                    log_content = f.read()

        # make sure print still appears in output
        self.assertIn("running my_task", output)

        # check that log messages don't go into output
        self.assertNotIn("Starting Locust", output)
        self.assertNotIn("Run time limit set to 1 seconds", output)

        # check that log messages goes into file
        self.assertIn(
            "%s/INFO/locust.main: Run time limit set to 1 seconds" % socket.gethostname(),
            log_content,
        )
        self.assertIn(
            "%s/INFO/locust.main: Time limit reached. Stopping Locust." % socket.gethostname(),
            log_content,
        )
        self.assertIn(
            "%s/INFO/locust.main: Shutting down (exit code 0), bye." % socket.gethostname(),
            log_content,
        )
        # check that message of custom logger also went into log file
        self.assertIn(
            "%s/INFO/root: custom log message" % socket.gethostname(),
            log_content,
        )



from locust.user.wait_time import constant
from typing import Any, Callable, Dict, List, TypeVar, Union
from gevent import GreenletExit, greenlet
from gevent.pool import Group
from locust.clients import HttpSession
from locust.exception import LocustError, StopUser
from locust.util import deprecation
from .task import (
    TaskSet,
    DefaultTaskSet,
    get_tasks_from_base_classes,
    LOCUST_STATE_RUNNING,
    LOCUST_STATE_WAITING,
    LOCUST_STATE_STOPPING,
)


class UserMeta(type):
    """
    Meta class for the main User class. It's used to allow User classes to specify task execution
    ratio using an {task:int} dict, or a [(task0,int), ..., (taskN,int)] list.
    """

    def __new__(mcs, classname, bases, class_dict):
        # gather any tasks that is declared on the class (or it's bases)
        tasks = get_tasks_from_base_classes(bases, class_dict)
        class_dict["tasks"] = tasks

        if not class_dict.get("abstract"):
            # Not a base class
            class_dict["abstract"] = False

        deprecation.check_for_deprecated_task_set_attribute(class_dict)

        return type.__new__(mcs, classname, bases, class_dict)


class User(object, metaclass=UserMeta):
    """
    Represents a "user" which is to be spawned and attack the system that is to be load tested.

    The behaviour of this user is defined by its tasks. Tasks can be declared either directly on the
    class by using the :py:func:`@task decorator <locust.task>` on methods, or by setting
    the :py:attr:`tasks attribute <locust.User.tasks>`.

    This class should usually be subclassed by a class that defines some kind of client. For
    example when load testing an HTTP system, you probably want to use the
    :py:class:`HttpUser <locust.HttpUser>` class.
    """

    host: str = None
    """Base hostname to swarm. i.e: http://127.0.0.1:1234"""

    min_wait = None
    """Deprecated: Use wait_time instead. Minimum waiting time between the execution of locust tasks"""

    max_wait = None
    """Deprecated: Use wait_time instead. Maximum waiting time between the execution of locust tasks"""

    wait_time = constant(0)
    """
    Method that returns the time (in seconds) between the execution of locust tasks.
    Can be overridden for individual TaskSets.

    Example::

        from locust import User, between
        class MyUser(User):
            wait_time = between(3, 25)
    """

    wait_function = None
    """
    .. warning::

        DEPRECATED: Use wait_time instead. Note that the new wait_time method should return seconds and not milliseconds.

    Method that returns the time between the execution of locust tasks in milliseconds
    """

    tasks: List[Union[TaskSet, Callable]] = []
    """
    Collection of python callables and/or TaskSet classes that the Locust user(s) will run.

    If tasks is a list, the task to be performed will be picked randomly.

    If tasks is a *(callable,int)* list of two-tuples, or a {callable:int} dict,
    the task to be performed will be picked randomly, but each task will be weighted
    according to its corresponding int value. So in the following case, *ThreadPage* will
    be fifteen times more likely to be picked than *write_post*::

        class ForumPage(TaskSet):
            tasks = {ThreadPage:15, write_post:1}
    """

    weight = 1
    """Probability of user class being chosen. The higher the weight, the greater the chance of it being chosen."""

    abstract = True
    """If abstract is True, the class is meant to be subclassed, and locust will not spawn users of this class during a test."""

    environment = None
    """A reference to the :py:attr:`environment <locust.Environment>` in which this locust is running"""

    client = None
    _state = None
    _greenlet: greenlet.Greenlet = None
    _group: Group
    _taskset_instance = None

    def __init__(self, environment):
        super().__init__()
        self.environment = environment

    def on_start(self):
        """
        Called when a User starts running.
        """
        pass

    def on_stop(self):
        """
        Called when a User stops running (is killed)
        """
        pass

    def run(self):
        self._state = LOCUST_STATE_RUNNING
        self._taskset_instance = DefaultTaskSet(self)
        try:
            # run the TaskSet on_start method, if it has one
            self.on_start()

            self._taskset_instance.run()
        except (GreenletExit, StopUser):
            # run the on_stop method, if it has one
            self.on_stop()

    def wait(self):
        """
        Make the running user sleep for a duration defined by the User.wait_time
        function.

        The user can also be killed gracefully while it's sleeping, so calling this
        method within a task makes it possible for a user to be killed mid-task even if you've
        set a stop_timeout. If this behaviour is not desired, you should make the user wait using
        gevent.sleep() instead.
        """
        self._taskset_instance.wait()

    def start(self, group: Group):
        """
        Start a greenlet that runs this User instance.

        :param group: Group instance where the greenlet will be spawned.
        :type group: gevent.pool.Group
        :returns: The spawned greenlet.
        """

        def run_user(user):
            """
            Main function for User greenlet. It's important that this function takes the user
            instance as an argument, since we use greenlet_instance.args[0] to retrieve a reference to the
            User instance.
            """
            user.run()

        self._greenlet = group.spawn(run_user, self)
        self._group = group
        return self._greenlet

    def stop(self, force=False):
        """
        Stop the user greenlet.

        :param force: If False (the default) the stopping is done gracefully by setting the state to LOCUST_STATE_STOPPING
                      which will make the User instance stop once any currently running task is complete and on_stop
                      methods are called. If force is True the greenlet will be killed immediately.
        :returns: True if the greenlet was killed immediately, otherwise False
        """
        if force or self._state == LOCUST_STATE_WAITING:
            self._group.killone(self._greenlet)
            return True
        elif self._state == LOCUST_STATE_RUNNING:
            self._state = LOCUST_STATE_STOPPING
            return False

    @property
    def group(self):
        return self._group

    @property
    def greenlet(self):
        return self._greenlet

    def context(self) -> Dict:
        """
        Adds the returned value (a dict) to the context for :ref:`request event <request_context>`
        """
        return {}

    @classmethod
    def fullname(cls) -> str:
        """Fully qualified name of the user class, e.g. my_package.my_module.MyUserClass"""
        return ".".join(filter(lambda x: x != "<locals>", (cls.__module__ + "." + cls.__qualname__).split(".")))


class HttpUser(User):
    """
    Represents an HTTP "user" which is to be spawned and attack the system that is to be load tested.

    The behaviour of this user is defined by its tasks. Tasks can be declared either directly on the
    class by using the :py:func:`@task decorator <locust.task>` on methods, or by setting
    the :py:attr:`tasks attribute <locust.User.tasks>`.

    This class creates a *client* attribute on instantiation which is an HTTP client with support
    for keeping a user session between requests.
    """

    abstract = True
    """If abstract is True, the class is meant to be subclassed, and users will not choose this locust during a test"""

    client: HttpSession = None
    """
    Instance of HttpSession that is created upon instantiation of Locust.
    The client supports cookies, and therefore keeps the session between HTTP requests.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.host is None:
            raise LocustError(
                "You must specify the base host. Either in the host attribute in the User class, or on the command line using the --host option."
            )

        session = HttpSession(
            base_url=self.host,
            request_event=self.environment.events.request,
            user=self,
        )
        session.trust_env = False
        self.client = session



import time
import unittest

from locust import User
from locust.distribution import weight_users


class TestDistribution(unittest.TestCase):
    def test_distribution_no_user_classes(self):
        user_classes_count = weight_users(user_classes=[], user_count=0)
        self.assertDictEqual(user_classes_count, {})

        user_classes_count = weight_users(user_classes=[], user_count=1)
        self.assertDictEqual(user_classes_count, {})

    def test_distribution_equal_weights_and_fewer_amount_than_user_classes(self):
        class User1(User):
            weight = 1

        class User2(User):
            weight = 1

        class User3(User):
            weight = 1

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=0)
        self.assertDictEqual(user_classes_count, {"User1": 0, "User2": 0, "User3": 0})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=1)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 0, "User3": 0})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=2)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 1, "User3": 0})

    def test_distribution_equal_weights(self):
        class User1(User):
            weight = 1

        class User2(User):
            weight = 1

        class User3(User):
            weight = 1

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=3)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 1, "User3": 1})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=4)
        self.assertDictEqual(user_classes_count, {"User1": 2, "User2": 1, "User3": 1})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=5)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 2, "User3": 2})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=6)
        self.assertDictEqual(user_classes_count, {"User1": 2, "User2": 2, "User3": 2})

    def test_distribution_unequal_and_unique_weights_and_fewer_amount_than_user_classes(self):
        class User1(User):
            weight = 1

        class User2(User):
            weight = 2

        class User3(User):
            weight = 3

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=0)
        self.assertDictEqual(user_classes_count, {"User1": 0, "User2": 0, "User3": 0})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=1)
        self.assertDictEqual(user_classes_count, {"User1": 0, "User2": 0, "User3": 1})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=2)
        self.assertDictEqual(user_classes_count, {"User1": 0, "User2": 1, "User3": 1})

    def test_distribution_unequal_and_unique_weights(self):
        class User1(User):
            weight = 1

        class User2(User):
            weight = 2

        class User3(User):
            weight = 3

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=3)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 1, "User3": 1})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=4)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 1, "User3": 2})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=5)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 2, "User3": 2})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=6)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 2, "User3": 3})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=10)
        self.assertDictEqual(user_classes_count, {"User1": 2, "User2": 3, "User3": 5})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=11)
        self.assertDictEqual(user_classes_count, {"User1": 2, "User2": 4, "User3": 5})

    def test_distribution_unequal_and_non_unique_weights_and_fewer_amount_than_user_classes(self):
        class User1(User):
            weight = 1

        class User2(User):
            weight = 2

        class User3(User):
            weight = 2

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=0)
        self.assertDictEqual(user_classes_count, {"User1": 0, "User2": 0, "User3": 0})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=1)
        self.assertDictEqual(user_classes_count, {"User1": 0, "User2": 1, "User3": 0})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=2)
        self.assertDictEqual(user_classes_count, {"User1": 0, "User2": 1, "User3": 1})

    def test_distribution_unequal_and_non_unique_weights(self):
        class User1(User):
            weight = 1

        class User2(User):
            weight = 2

        class User3(User):
            weight = 2

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=3)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 1, "User3": 1})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=4)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 1, "User3": 2})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=5)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 2, "User3": 2})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=6)
        self.assertDictEqual(user_classes_count, {"User1": 1, "User2": 3, "User3": 2})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=10)
        self.assertDictEqual(user_classes_count, {"User1": 2, "User2": 4, "User3": 4})

        user_classes_count = weight_users(user_classes=[User1, User2, User3], user_count=11)
        self.assertDictEqual(user_classes_count, {"User1": 2, "User2": 5, "User3": 4})

    def test_distribution_large_number_of_users(self):
        class User1(User):
            weight = 5

        class User2(User):
            weight = 55

        class User3(User):
            weight = 37

        class User4(User):
            weight = 2

        class User5(User):
            weight = 97

        class User6(User):
            weight = 41

        class User7(User):
            weight = 33

        class User8(User):
            weight = 19

        class User9(User):
            weight = 19

        class User10(User):
            weight = 34

        class User11(User):
            weight = 78

        class User12(User):
            weight = 76

        class User13(User):
            weight = 28

        class User14(User):
            weight = 62

        class User15(User):
            weight = 69

        for user_count in range(1044523783783, 1044523783783 + 1000):
            ts = time.perf_counter()
            user_classes_count = weight_users(
                user_classes=[
                    User1,
                    User2,
                    User3,
                    User4,
                    User5,
                    User6,
                    User7,
                    User8,
                    User9,
                    User10,
                    User11,
                    User12,
                    User13,
                    User14,
                    User15,
                ],
                user_count=user_count,
            )
            delta_ms = 1e3 * (time.perf_counter() - ts)
            self.assertEqual(sum(user_classes_count.values()), user_count)
            self.assertLessEqual(delta_ms, 100)
